/* grWin3.c -
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1985, 1990 Regents of the University of California. * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  The University of California        * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.  Export of this software outside     * 
 *     * of the United States of America may require an export license.    * 
 *     *********************************************************************
 *
 * This file contains additional functions to manipulate an X window system
 * color display.  Included here are device-dependent routines to draw and 
 * erase text and draw a grid.
 *
 */

#include <stdio.h>
#include <w32api/windows.h>
#include <w32api/windowsx.h>

#include "misc/magic.h"
#include "utils/geometry.h"
#include "graphics/graphics.h"
#include "windows/windows.h"
#include "graphics/graphicsInt.h"
#include "textio/textio.h"
#include "signals/signals.h"
#include "utils/utils.h"
#include "utils/hash.h"
#include "grWinInt.h"

/* locals */

static HFONT grWinFonts[4];
#define	grSmallFont	grWinFonts[0]
#define	grMediumFont	grWinFonts[1]
#define	grLargeFont	grWinFonts[2]
#define	grXLargeFont	grWinFonts[3]



/*---------------------------------------------------------
 * grwinDrawGrid:
 *	grxDrawGrid adds a grid to the grid layer, using the current
 *	write mask and color.
 *
 * Results:
 *	TRUE is returned normally.  However, if the grid gets too small
 *	to be useful, then nothing is drawn and FALSE is returned.
 *
 * Side Effects:    None.
 *---------------------------------------------------------
 */

bool
grwinDrawGrid (prect, outline, clip)
    Rect *prect;			/* A rectangle that forms the template
			         * for the grid.  Note:  in order to maintain
			         * precision for the grid, the rectangle
			         * coordinates are specified in units of
			         * screen coordinates multiplied by 4096.
			         */
    int outline;		/* the outline style */
    Rect *clip;			/* a clipping rectangle */
{
    int xsize, ysize;
    int x, y;
    int xstart, ystart;
    int low, hi, shifted, snum;

    xsize = prect->r_xtop - prect->r_xbot;
    ysize = prect->r_ytop - prect->r_ybot;
    if (GRID_TOO_SMALL(xsize, ysize, GR_NUM_GRIDS)) return FALSE;
    
    xstart = prect->r_xbot % xsize;
    while (xstart < clip->r_xbot<<12) xstart += xsize;
    ystart = prect->r_ybot % ysize;
    while (ystart < clip->r_ybot<<12) ystart += ysize;
    
    grwinSetLineStyle(outline);
    SelectPen(hdc, grPenDraw);

    snum = 0;
    low = grMagicToWin(clip->r_ybot);
    hi = grMagicToWin(clip->r_ytop);
    for (x = xstart; x < (clip->r_xtop+1)<<12; x += xsize)
    {
        if (snum >= GR_NUM_GRIDS) break;
	shifted = x >> 12;
	MoveToEx(hdc, shifted, low, NULL);
	LineTo(hdc, shifted, hi);
    }

    snum = 0;
    low = clip->r_xbot;
    hi = clip->r_xtop;
    for (y = ystart; y < (clip->r_ytop+1)<<12; y += ysize)
    {
        if (snum >= GR_NUM_GRIDS) break;
	shifted = grMagicToWin(y >> 12);
	MoveToEx(hdc, low, shifted, NULL);
	LineTo(hdc, hi, shifted);
    }
    return true;
}


/*---------------------------------------------------------
 * grwinLoadFont
 *	This local routine loads the X fonts used by Magic.
 *
 * Results:	None.
 *
 * Side Effects:    None.
 *---------------------------------------------------------
 */

Void
grwinLoadFont()
{
    static int defaultfontsizes[4] = {
	/* 10, 14, 18, 24 */
	18, 24, 30, 40		/* These are pixel heights */
    };

    int i;

    for (i = 0; i != 4; i++)
    {
	/* Load resources here. . . */

	/* otherwise load Helvetica (Arial) medium-upright-normal */

	grWinFonts[i] = CreateFont(defaultfontsizes[i], 0, 0, 0,
		FW_MEDIUM, 0, 0, 0, ANSI_CHARSET,
		OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		DEFAULT_PITCH | FF_SWISS, "Arial");
    }
    return true;
}


/*---------------------------------------------------------
 * grwinSetCharSize:
 *	This local routine sets the character size in the display,
 *	if necessary.
 *
 * Results:	None.
 *
 * Side Effects:    None.
 *---------------------------------------------------------
 */

Void
grwinSetCharSize (size)
    int size;		/* Width of characters, in pixels (6 or 8). */
{
    grCurrent.fontSize = size;
    switch (size)
    {
	case GR_TEXT_DEFAULT:
	case GR_TEXT_SMALL:
	    grCurrent.font = grSmallFont;
	    break;
	case GR_TEXT_MEDIUM:
	    grCurrent.font = grMediumFont;
	    break;
	case GR_TEXT_LARGE:
	    grCurrent.font = grLargeFont;
	    break;
	case GR_TEXT_XLARGE:
	    grCurrent.font = grXLargeFont;
	    break;
	default:
	    TxError("%s%d\n", "grwinSetCharSize: Unknown character size ",
		size );
	    break;
    }
    SelectFont(hdc, grCurrent.font);
}


/*
 * ----------------------------------------------------------------------------
 * GrWinTextSize --
 *
 *	Determine the size of a text string. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A rectangle is filled in that is the size of the text in pixels.
 *	The origin (0, 0) of this rectangle is located on the baseline
 *	at the far left side of the string.
 * ----------------------------------------------------------------------------
 */

Void
GrWinTextSize(text, size, r)
    char *text;
    int size;
    Rect *r;
{
    TEXTMETRIC textinfo;
    SIZE stringsize;
    
    if (grCurrent.fontSize != size)
    {
	switch (size) {
	case GR_TEXT_DEFAULT:
	case GR_TEXT_SMALL:
	    SelectFont(hdc,grSmallFont);
	    break;
	case GR_TEXT_MEDIUM:
	    SelectFont(hdc,grMediumFont);
	    break;
	case GR_TEXT_LARGE:
	    SelectFont(hdc,grLargeFont);
	    break;
	case GR_TEXT_XLARGE:
	    SelectFont(hdc,grXLargeFont);
	    break;
	default:
	    TxError("%s%d\n", "GrWinTextSize: Unknown character size ",
		size );
	    break;
	}
    }
    GetTextMetrics(hdc, &textinfo);
    GetTextExtentPoint32(hdc, text, strlen(text), &stringsize);

    r->r_ytop = textinfo.tmAscent;
    r->r_ybot = -textinfo.tmDescent;
    r->r_xtop = stringsize.cx;
    r->r_xbot = 0;
}


/*
 * ----------------------------------------------------------------------------
 * GrWinReadPixel --
 *
 *	Read one pixel from the screen.
 *
 * Results:
 *	An integer containing the pixel's color.
 *
 * Side effects:
 *	none.
 *
 * ----------------------------------------------------------------------------
 */

int
GrWinReadPixel (w, x, y)
    MagWindow *w;
    int x,y;		/* the location of a pixel in screen coords */
{
    COLORREF loccolor;
    /* To be done */
    return 0;
}


/*
 * ----------------------------------------------------------------------------
 * GrWinBitBlt --
 *
 *	Copy information from one part of the screen to the other.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	changes the screen.
 * ----------------------------------------------------------------------------
 */

Void
GrWinBitBlt(r, p)
    Rect *r;
    Point *p;
{
    BitBlt(hdc, r->r_xbot, grMagicToWin(r->r_ybot),
		r->r_xtop - r->r_xbot + 1, r->r_ytop - r->r_ybot + 1,
		hdc,
		p->p_x, grMagicToWin(p->p_y),
		SRCCOPY);
}



/*
 * ----------------------------------------------------------------------------
 * GrWinRectConvert --
 *	Convert a magic rectangle into an Windows RECT
 *	(Both passed as pointers)
 * ----------------------------------------------------------------------------
 */

Void
grwinRectConvert(mr, wr)
    Rect *mr;
    RECT *wr;
{
    wr->left = mr->r_xbot;
    wr->bottom = grMagicToWin(mr->r_ybot);
    wr->right = mr->r_xtop;
    wr->top = grMagicToWin(mr->r_ytop);
}

/*---------------------------------------------------------
 * grwinPutText:
 *
 *	This routine puts a chunk of text on the screen in the current
 *	color, size, etc.  The caller must ensure that it fits on
 *	the screen -- no clipping is done except to the obscuring rectangle
 *	list and the clip rectangle.
 *
 * Results:	
 *	none.
 *
 * Side Effects:
 *	The text is drawn on the screen.  
 *
 *---------------------------------------------------------
 */

Void
grwinPutText (text, pos, clip, obscure)
    char *text;			/* The text to be drawn. */
    Point *pos;			/* A point located at the leftmost point of
				 * the baseline for this string.
				 */
    Rect *clip;			/* A rectangle to clip against */
    LinkedRect *obscure;	/* A list of obscuring rectangles */

{
    Rect location;
    Rect overlap;
    Rect textrect;
    LinkedRect *ob;
    void grWinGeoSub();
    HRGN clipregion;

    if (grCurrent.font == NULL) return;

    GrWinTextSize(text, grCurrent.fontSize, &textrect);

    location.r_xbot = pos->p_x + textrect.r_xbot;
    location.r_xtop = pos->p_x + textrect.r_xtop;
    location.r_ybot = pos->p_y + textrect.r_ybot;
    location.r_ytop = pos->p_y + textrect.r_ytop;

    /* erase parts of the bitmap that are obscured */
    for (ob = obscure; ob != NULL; ob = ob->r_next)
    {
	if (GEO_TOUCH(&ob->r_r, &location))
	{
	    overlap = location;
	    GeoClip(&overlap, &ob->r_r);
	    grWinGeoSub(&location, &overlap);
	}
    }
 
    overlap = location;
    GeoClip(&overlap, clip);

    /* copy the text to the color screen */
    if ((overlap.r_xbot < overlap.r_xtop)&&(overlap.r_ybot <= overlap.r_ytop))
    {
	RECT wr;

	SelectFont(hdc, grCurrent.font);
	grwinRectConvert(&overlap, &wr);
	clipregion = CreateRectRgnIndirect(&wr);
	SelectClipRgn(hdc, clipregion);
	SelectBrush(hdc, grBrushText);	
	BeginPath(hdc);
	SetBkMode(hdc, TRANSPARENT);
	TextOut(hdc, pos->p_x,
		grMagicToWin(pos->p_y + textrect.r_ytop),
		text, strlen(text));
	EndPath(hdc);
	FillPath(hdc);
	SelectClipRgn(hdc, NULL);	/* reset the clipping region */
    }
}


/* grWinGeoSub:
 *	return the tallest sub-rectangle of r not obscured by area
 *	area must be within r.
 */

void
grWinGeoSub(r, area)
register Rect *r;		/* Rectangle to be subtracted from. */
register Rect *area;		/* Area to be subtracted. */

{
    if (r->r_xbot == area->r_xbot) r->r_xbot = area->r_xtop;
    else
    if (r->r_xtop == area->r_xtop) r->r_xtop = area->r_xbot;
    else
    if (r->r_ybot <= area->r_ybot) r->r_ybot = area->r_ytop;
    else
    if (r->r_ytop == area->r_ytop) r->r_ytop = area->r_ybot;
    else
    r->r_xtop = area->r_xbot;
}
